{
  "type": "record",
  "name": "Status",
  "namespace": "com.linkedin.pegasus2avro.common",
  "doc": "The status metadata of an entity, e.g. dataset, metric, feature, etc.\nThis aspect is used to represent soft deletes conventionally.",
  "fields": [
    {
      "name": "removed",
      "type": "boolean",
      "doc": "whether the entity is removed or not",
      "default": false,
      "Searchable": {
        "fieldType": "BOOLEAN"
      }
    }
  ],
  "Aspect": {
    "name": "status"
  }
}