# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import GlossaryNodeInfoClass
from .....schema_classes import GlossaryRelatedTermsClass
from .....schema_classes import GlossaryTermInfoClass


GlossaryNodeInfo = GlossaryNodeInfoClass
GlossaryRelatedTerms = GlossaryRelatedTermsClass
GlossaryTermInfo = GlossaryTermInfoClass
# fmt: on
