{
  "type": "record",
  "name": "SubTypes",
  "namespace": "com.linkedin.pegasus2avro.common",
  "doc": "Sub Types. Use this aspect to specialize a generic Entity\ne.g. Making a Dataset also be a View or also be a LookerExplore",
  "fields": [
    {
      "name": "typeNames",
      "type": {
        "type": "array",
        "items": "string"
      },
      "doc": "The names of the specific types.\n"
    }
  ],
  "Aspect": {
    "name": "subTypes"
  }
}